// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/* written by John Doug Reynolds, June 2002 */

#include "calib.h"
#include <CalServer.h>
#include <Convolver.h>
#include <errstr.h>
#include <algorithm>
#include <vector>

using namespace scanningLib;

static vector<float>& wrap( vector<float>&, const Spectrum& );
static void shift( vector<float>&, const Spectrum&, int offset );
static void rebin( struct Spectrum &vec, float min, float max, float val );

static double calBeta( double phi /* arcmin, dispersion off-axis angle */,
		       int order, double energy /* eV */ );

static Spectrum customFigureDistribution( double betaBinWidth,
					  int order, double energy /* eV */ );

static CcdQuantumEfficiencyCurve	ccdqeCurve[9];
static EffectiveAreaDataServer*		caldsEA;
static EffectiveAreaCurve*		eaCurve[6];
static GeometryDataServer*		caldsGM;
static GratingDataServer*		caldsGR;
static PsfDataServer*			caldsPS;
static GridSpan				betaGrid;
static double				theta, phi;

vector<AngularDistBin> customFigure;
vector<double> vignetting;

/* --- LSF components ------------------------------------------------------ */

#if 1
void prfBroad( vector<float> &rf,
	       float eLow, float eHigh, float alpha, int order ) 
{
   double energy = 1000 * (eLow + eHigh) / 2;
   double beta_m = caldsGR->beta(alpha,order,energy);
   int curveHead = betaGrid.cell(beta_m) - betaGrid.min;
   int right = max( (int) betaGrid.length() - curveHead - 1, 0 );
   int left = max( curveHead, 0 );

   Spectrum
      dist=caldsGR->scatteringDistribution(left, right, betaGrid.stride,
					   alpha, order, energy, false, true);
   shift( rf, dist, curveHead - left );

   float effective_area = eaCurve[-order]->area(1000*eLow,1000*eHigh);
   for ( unsigned i = 0; i < betaGrid.length(); ++i ) rf[i] *= effective_area;
}
#else
void prfBroad( vector<float> &rf,
	       float eLow, float eHigh, float alpha, int order ) 
{
   double energy = 1000 * (eLow + eHigh) / 2;
   double beta_m = caldsGR->beta(alpha,order,energy);

   Spectrum
      dist=caldsGR->scatteringDistribution(100, betaGrid.stride,
					   alpha, order, energy, false, true);

   shift( rf, dist, betaGrid.cell(beta_m)-betaGrid.min-(int)dist.size()/2 );

   float effective_area = eaCurve[-order]->area(1000*eLow,1000*eHigh);
   for ( unsigned i = 0; i < betaGrid.length(); ++i ) rf[i] *= effective_area;
}
#endif

void prfNarrow( vector<float> &rf,
		float eLow, float eHigh, float alpha, int order,
		int fftdim, bool withmirrorpsf, bool withangdist )
{
   double dBeta = betaGrid.stride;
   double energy = 1000 * (eLow + eHigh) / 2;
   double beta_m = caldsGR->beta(alpha,order,energy);
   struct Spectrum dist;

   if ( betaGrid.contains( betaGrid.cell(beta_m) ) ) {
      list<Spectrum> kernels;
      double cm = sin(alpha) / sin(beta_m); // chromatic magnification
      double nhwhm;

      // small angle scattering

      if ( fftdim == 1 ) nhwhm = 1.5;
      else if ( fftdim == 2 ) nhwhm = 3;
      else nhwhm = 4;

      dist = caldsGR->scatteringDistribution(nhwhm,dBeta,
					     alpha,order,energy,true,false);

      DPair fSmallLarge = caldsGR->scatteringDistributionContribs();
      dist[dist.size()/2] += (1 - fSmallLarge.first)*(1 - fSmallLarge.second);
      kernels.push_back( dist );

      // mirror point spread contribution

      if ( withmirrorpsf ) {
	 if ( fftdim == 1 ) nhwhm = 36;
	 else if ( fftdim == 2 ) nhwhm = 72;
	 else if ( fftdim == 3 ) nhwhm = 157;
	 else if ( fftdim == 4 ) nhwhm = 340;
	 else nhwhm = 706;

	 kernels.push_back
	    ( caldsPS->dispersionFigureDistribution(nhwhm,dBeta,cm) );
      }

      // custom angular distribution

      if ( withangdist )
	 kernels.push_back
	    ( customFigureDistribution(dBeta,order,energy) );

      // grating bow induced misalignment

      kernels.push_back
	 ( caldsGR->bowingFigureDistribution(8,dBeta,cm) );

      // geometric defocus approximation

      kernels.push_back
	 ( caldsGM->lsfDefocusDistribution(1,dBeta,theta,phi,order,energy) );

      // finite energy band contribution

      double dLambda = hc/eLow - hc/eHigh;
      kernels.push_back
	 ( caldsGR->broadeningDistribution(dBeta,dLambda,beta_m,order) );

      // grating misalignment

      double ico = betaGrid.tick(beta_m) - betaGrid.cell(beta_m);
      dist = caldsGR->misalignmentFigureDistribution(8,dBeta,cm,ico);

      // convolve grating misalignment against the accumulated kernels

      list<Spectrum>::const_iterator itr;

      unsigned maxKernelSize = 0;
      for ( itr = kernels.begin(); itr != kernels.end(); ++itr )
	 maximize( maxKernelSize, itr->size() );

      unsigned size = dist.size() + maxKernelSize;
      unsigned padding = (1 + maxKernelSize) / 2;

      struct Convolver
       convolver( 1 << (int) ceil( 
	static_cast<long double>(std::log(static_cast<long double>(size+padding))) /
        static_cast<long double>(std::log(static_cast<long double>(2))) ) );

      vector<float> curve( convolver.size() );
      shift( curve, dist, (int)size/2 - (int)dist.size()/2 );
      convolver.load( curve );

      for ( itr = kernels.begin(); itr != kernels.end(); ++itr )
	 if ( itr->size() ) convolver.load( wrap(curve,*itr) );

      dist.resize( size );
      dist.assign( convolver().begin(), convolver().begin() + dist.size() );
   }
   else
      dist = caldsGR->scatteringDistribution(3,dBeta,
					     alpha,order,energy,true,false);

   shift( rf, dist, betaGrid.cell(beta_m)-betaGrid.min-(int)dist.size()/2 );

   double effective_area = eaCurve[-order]->area(1000*eLow,1000*eHigh);
   for ( unsigned i = 0; i < betaGrid.length(); ++i ) rf[i] *= effective_area;
}

/* --- calibrations -------------------------------------------------------- */

float empiricalCrossLA( float energy, float dpsi, float xdsp0, float xdsp1 )
{
   float k = (energy * 10800) / (0.678 * M_PI);
   return (atan(k*(xdsp1+dpsi)) - atan(k*(xdsp0+dpsi))) / M_PI;
}

double calBeta( double phi /* arcmin */, int order, double energy /* eV */ )
{
   return caldsGR->beta( caldsGR->alpha(-60*phi,0), order, energy );
}

float calEnergy( float beta, int order, float alpha )
{
   return caldsGR->energy(alpha,beta,order);
}

float calAlpha0()
{
   return caldsGR->alpha0();
}

void ccd_qe( float qe[10], float energy /* keV */ )
{
   qe[0] = 0;
   for ( unsigned i = 1; i <= 9; ++i )
      qe[i] = ccdqeCurve[i-1].efficiency( energy * 1000 );
}

Spectrum customFigureDistribution( double betaBinWidth,
				   int order, double energy )
{
   struct Grid grid( betaBinWidth, calBeta(0,order,energy) );

   double betaMin = calBeta( customFigure.front().phi0, order, energy );
   double betaMax = calBeta( customFigure.back().phi1,  order, energy );

   int width = max( grid.cell(betaMax), -grid.cell(betaMin) );

   struct Spectrum dist( 1 + 2*width );
   grid.origin -= width * grid.stride;

   vector<AngularDistBin>::const_iterator itr;
   for ( itr = customFigure.begin(); itr != customFigure.end(); ++itr ) {
      float idx0 = 0.5 + grid.tick( calBeta(itr->phi0,order,energy) );
      float idx1 = 0.5 + grid.tick( calBeta(itr->phi1,order,energy) );
      rebin( dist, idx0, idx1, itr->value );
   }

   return dist;
}

/* --- initialize/finalize ------------------------------------------------- */

void calibInit( const GridSpan &_betaGrid )
{
   betaGrid = _betaGrid;

   caldsGR = 0; caldsGR = calServer.getAtom(caldsGR);
   caldsEA = 0; caldsEA = calServer.getAtom(caldsEA);
   caldsPS = 0; caldsPS = calServer.getAtom(caldsPS);
   caldsGM = 0; caldsGM = calServer.getAtom(caldsGM);

   unsigned i;
   for ( i = 0; i < 6; ++i ) eaCurve[i] = 0;

   for ( i = 0; i < 9; ++i ) {
      struct CcdQuantumDataServer *qds=0;
      calServer.state()->ccd()->set(i+1);
      qds = calServer.getAtom(qds);
      ccdqeCurve[i] = qds->grade(1);
      calServer.releaseAtom(qds);
   }

   vector<double> beta(betaGrid.length());
   for ( i=0; i < betaGrid.length(); ++i ) beta[i] = betaGrid[i+betaGrid.min];
   vignetting = caldsGR->selfVignettingEfficiency(beta);
}

void resetSourcePosition( float _theta, float _phi )
{
   theta = _theta;
   phi = _phi;
   for ( unsigned i = 0; i < 6; ++i ) {
      delete eaCurve[i];
      eaCurve[i] = new EffectiveAreaCurve(
			caldsEA->intrinsicEffectiveAreaCurve(-i,theta,phi) );
   }
}

void calibFree()
{
   for ( unsigned i = 0; i < 6; ++i ) delete eaCurve[i];
   calServer.releaseAtom(caldsGM);
   calServer.releaseAtom(caldsPS);
   calServer.releaseAtom(caldsEA);
   calServer.releaseAtom(caldsGR);
}

/* --- utility ------------------------------------------------------------- */

vector<float>& wrap( vector<float> &curve, const Spectrum &dist )
{
   vector<double>::const_iterator center, begin, end, itr;

   if ( dist.size() > curve.size() ) {
      begin = dist.begin() + (dist.size() - curve.size())/2;
      end = begin + curve.size();
   }
   else {
      begin = dist.begin();
      end = dist.end();
   }

   center = dist.begin() + dist.size()/2;
   if ( center < begin )
      center = begin;
   else if ( center > end )
      center = end;

   float *ptr = &curve.front();
   for ( itr = center; itr != end; *ptr++ = *itr++ );

   unsigned count = curve.size() - (end-begin);
   memset( ptr, 0, count * sizeof(float) );
   ptr += count;

   for ( itr = begin; itr != center; *ptr++ = *itr++ );

   return curve;
}

void shift( vector<float> &curve, const Spectrum &dist, int offset )
{
   float *ptr = &curve.front();

   int begin = max( offset, 0 );
   int end = min( (int)dist.size() + offset, (int)curve.size() );

   if ( end < begin ) {
      memset( ptr, 0, curve.size() * sizeof(float) );
      return;
   }

   if ( begin > 0 ) {
      memset( ptr, 0, begin * sizeof(float) );
      ptr += begin;
   }

   vector<double>::const_iterator itr, endI = dist.begin() + end - offset;
   for ( itr = dist.begin() + begin - offset; itr != endI; *ptr++ = *itr++ );

   if ( (int)curve.size() > end )
      memset( ptr, 0, (curve.size() - end) * sizeof(float) );
}

void rebin( struct Spectrum &vec, float min, float max, float val )
{
   if ( max < min ) swap(min,max);

   if ( min >= vec.size() || max <= 0 ) return;

   if ( min < 0 ) {
      val *= 1 + min/(max - min);
      min = 0;
   }

   if ( max > vec.size() ) {
      val *= 1 - (max - vec.size())/(max - min);
      max = vec.size();
   }

   if ( floor(max) > floor(min) ) {
      float portion = val / (max - min);
      if ( ceil(min) > min ) {
	 vec[int(floor(min))] += portion * (ceil(min) - min);
	 min = ceil(min);
      }
      while ( floor(max) > floor(min) ) {
	 vec[int(floor(min))] += portion;
	 min += 1;
      }
      vec[int(floor(min))] += portion * (max - min);
   }
   else
      vec[int(floor(min))] += val;
}
