// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/* written by John Doug Reynolds, June 2002 */

#include "rgsrmfgen.h"
#include "calib.h"
#include <ShapeScanner.h>
#include <rgsPixelsLib.h>
#include <DalVec.h>
#include <ASCregion.h>
#include <CalServer.h>
#include <cstdio>
#include "param.h"

using namespace rgsPixelsLib;
using namespace std;

struct LossStats {
	list<Span>			spans;		// extraction column segments
	float			exposure;	// normalized exposure
	float			crossPSF;	// normalized exposure*crossPSF
	void ingest( const ExposureMap&, const Spectrum&, int column, int offset );
	void normalize( float norm ) { exposure *= norm; crossPSF *= norm; }
	LossStats() : exposure(0), crossPSF(0) {}
};

struct LossFunction {
	unsigned char		ccd;		// CCD identifier
	unsigned			offset;		// beta offset of array origin
	vector<LossStats>		fg, bg;		// indexed by beta subspace
};

bool subdivides( const GridSpan &fine, const GridSpan &coarse );
struct ShapeScanner* scanner( struct Table*,
		const Grid &xGrid, const Grid &yGrid );

void init( const Table*, int sourceid, float &alpha, float &dpsi, int order );
void init( vector< list<Span> > &banana, struct ShapeScanner& );
void init( list<LossFunction>&, struct DataSet*, struct ShapeScanner &fg,
		struct ShapeScanner *bg, float dpsi, int order, string energyBinning );

float empiricalCrossLA( float energy, float dpsi, const list<Span>& );

void nodeEfficiencySA( vector<float>&, float energy);
void nodeEfficiencyLA( vector<float>&, float dpsi, float energy);
void betaEfficiency( vector<float>&, float energy);

std::vector< list<Span> >	banana;
list<LossFunction>	lossByNode;
struct GridSpan		betaGrid, xdspGrid;
//AI: Wavelength Scale. Start
struct GridSpan betaChannelGrid;
struct GridSpan betaORIChannelGrid;

double alpha0, linespacing;
double offaxis;
bool _lambdaSpec;
//AI: Wavelength Scale. End
bool			betaEfficInit;


void rgsrmfgen()
{
	float low_threshold = 1e-7; // smallest non-zero response
	betaEfficInit = false;

	// process the task parameters

	bool withspectrum=false, bkgcorrect=false;
	bool withmirrorpsf=false, withangdist=false;
	string nameTemplate, spectrumset, angdistset;
	unsigned source=0;
	int order=0;
	bool witharffile=false;

	string evlist = stringParameter("evlist");
	string rmfset = stringParameter("rmfset");
	string arfset= stringParameter("arfset");
	string srclist = stringParameter("srclist");
	float emax = realParameter("emax");
	float emin = realParameter("emin");
	int rows = intParameter("rows");
	int fftdim = intParameter("fftdim");
	bool newrmf = booleanParameter("newrmf");
	std::string energyBinning = stringParameter("spectrumbinning");
	bool withrectification = booleanParameter("withrectification");
	bool withEffAreaCorr = booleanParameter("witheffectiveareacorrection");
	//AI: Wavelength scale.
	//Set it false by default to allow the possibility of
	//running rgsrmfgen with the parameter
	//withspectrumset=no
	_lambdaSpec = false;

	if ( newrmf ) {
		withspectrum = booleanParameter("withspectrum");
		witharffile = booleanParameter("witharffile");
		if (witharffile)
			 low_threshold = 1e-7;

		if ( withspectrum ) {
			spectrumset = stringParameter("spectrumset");
			if ( spectrumset != "" )
				nameTemplate = spectrumset;
			else
				errstr << "in this configuration --spectrumset must be explicit"
				<< Sas::error("NullFileName");
		}
		else {
			source = (unsigned) intParameter("source");
			order = intParameter("order");
			bkgcorrect = booleanParameter("bkgcorrect");
			if ( evlist != "" )
				nameTemplate = evlist;
			else if ( srclist != "" )
				nameTemplate = srclist;
			else
				errstr << "in this configuration either "
				<< "--evlist or --srclist must be explicit"
				<< Sas::error("NullFileName");
		}
		withmirrorpsf = booleanParameter("withmirrorpsf");
		withangdist = booleanParameter("withangdist");
		if ( withangdist )
			angdistset = stringParameter("angdistset");
	}
	else if ( rmfset != "" )
		nameTemplate = rmfset;
	else
		errstr << "in this configuration --rmfset must be explicit"
		<< Sas::error("NullFileName");

	string::size_type suffix = nameTemplate.rfind('.');
	if ( suffix == string::npos ) suffix = nameTemplate.size();

	if ( srclist == "" )
	{
		if ( suffix >= 10 )
			(srclist=nameTemplate).replace( suffix-10, 10, "SRCLI_0000" );
		else
			srclist = string("srcli") + nameTemplate.substr(suffix);
	}

	if ( evlist == "" )
	{
		if ( suffix >= 10 )
			(evlist=nameTemplate).replace( suffix-10, 10, "EVENLI0000" );
		else
			evlist = string("evenli") + nameTemplate.substr(suffix);
	}
	// dynamic parameter initializations

	struct DataSet* evenliDS = dataSetServer->open( evlist, Dal::Read );
	struct Table* etab = evenliDS->table("EVENTS");

	xdspGrid = gridspan( etab->column("XDSP_CHANNEL") );
	//AI: lambda binning
	betaORIChannelGrid = gridspan( etab->column("BETA_CHANNEL") );
	if (energyBinning == "beta")
		betaChannelGrid = gridspan( etab->column("BETA_CHANNEL") );     //AI: Wavelength scale. Start
	else
		betaChannelGrid = gridspan( etab->column("MLAMBDA_CHANNEL") );


	//AI: lambda binning

	struct DataSet* rmfDS;
	//ARF pointer initialization.
	struct DataSet* arfDS;
	real32 *arfD, *arfELD, *arfEHD;
	ColumnData *arfELCD, *arfEHCD, *arfCD;

	if ( newrmf )
	{
		if ( withspectrum )
		{
			struct DataSet* ds = dataSetServer->open( spectrumset, Dal::Read );
			const Table* table = ds->table("SPECTRUM");

			string hduclas2 = table->attribute("HDUCLAS2")->asString();
			if ( hduclas2 == "BKG" )
				errstr << "cannot make a response for this kind of spectrum (BKG)"
				<< Sas::error("InvalidSpectrum");
			else
				bkgcorrect = (hduclas2 == "NET");


			source = table->attribute("SOURCEID")->asInt();
			order  = table->attribute("RFLORDER")->asInt();
			//AI: Wavelength scale. Start
			betaGrid = gridspan( table->column("CHANNEL") );

			struct GridSpan chnGrid;
			if (table->column("CHANNEL")->attribute("TCUNI")->asString() == "Angstrom")
			{
				errstr << "Response matrix will be calculated in lambda space"
				<< Sas::message(Msg::App,Msg::Verbose);
				_lambdaSpec	= true;
				chnGrid = gridspan( etab->column("MLAMBDA_CHANNEL") );
			}
			else
				chnGrid = gridspan( etab->column("BETA_CHANNEL") );


			if ( !subdivides(chnGrid,betaGrid) )
				errstr << "SPECTRUM beta channels not compatible with EVENTS"
				<< Sas::error("InvalidRebinning");

			//AI: betaChannelGrid (BETA_CHANNEL in the event file) must have the same size (channels) that the
			//CHANNELS in the spectrum. If the spectrum was generated using rebin not equal to one, we have to
			//rebin the betaChannelGrid event channels to match the channels in the spectrum
			//AI: lambda binning
			size_t rebin;

			if  (energyBinning == "lambda")
			{
				rebin = (size_t) std::ceil((double) etab->column("MLAMBDA_CHANNEL")->attribute("TLMAX")->asInt()
					/ (double) table->column("CHANNEL")->attribute("TLMAX")->asInt());
			}
			else
			{
				rebin = (size_t) std::ceil((double) etab->column("BETA_CHANNEL")->attribute("TLMAX")->asInt()
									/ (double) table->column("CHANNEL")->attribute("TLMAX")->asInt());
			}


			size_t evnbins = (size_t) ceil( betaChannelGrid.length() / (float)rebin );


			betaChannelGrid = GridSpan( Grid( rebin*betaChannelGrid.stride, betaChannelGrid[(.5-betaChannelGrid.min)*(rebin-1)] ),
					Span(betaChannelGrid.min).length(evnbins) );

			//Same for BETA_CHANNEL
			 evnbins = (size_t) ceil( betaORIChannelGrid.length() / (float)rebin );
						betaORIChannelGrid = GridSpan( Grid( rebin*betaORIChannelGrid.stride, betaORIChannelGrid[(.5-betaORIChannelGrid.min)*(rebin-1)] ),
								Span(betaORIChannelGrid.min).length(evnbins) );


			//AI: lambda binning

			dataSetServer->close( ds );
		}
		else
		{
			if ( source == 0 )
				source = etab->attribute("SOURCEID")->asInt();
			if (_lambdaSpec)
				betaGrid = gridspan( etab->column("MLAMBDA_CHANNEL") );
			else
				betaGrid = gridspan( etab->column("BETA_CHANNEL") );
		}
		//AI: Wavelength scale. End

		if (arfset == "")
		{
			if (witharffile)
			{
				char scratch[32];
				std::sprintf( scratch, "ARFMAT%d%03d", order, source );
				(arfset=nameTemplate).replace( suffix-10, 10, scratch );
			}
		}

		if ( rmfset == "" )
		{
			if ( suffix >= 10 )
			{
				char scratch[32];
				if (!witharffile)
					std::sprintf( scratch, "RSPMAT%d%03d", order, source );
				else
					std::sprintf( scratch, "RMFMAT%d%03d", order, source );

				(rmfset=nameTemplate).replace( suffix-10, 10, scratch );

			}
			else
				rmfset = string("rspmat") + nameTemplate.substr(suffix);
		}


		enum Xmm::DataMode datamode
		= Xmm::toDataMode( evenliDS->attribute("DATAMODE")->asString() );
		if ( datamode == Xmm::HIGH_TIME_RESOLUTION ) bkgcorrect = false;

		rmfDS = dataSetServer->open( rmfset, Dal::Create );
		if (witharffile)
			arfDS = dataSetServer->open( arfset, Dal::Create);


		*rmfDS->addAttribute( "SOURCEID", "SRCLIST:INDEX source identifier" )
		= (int)source;
		*rmfDS->addAttribute( "RFLORDER", "ordinal reflection order" )
		= order;
		*rmfDS->addAttribute( "BKG_CORR", "background-corrected response" )
		= bkgcorrect;
		*rmfDS->addAttribute( "MIRRORPS", "the standard mirror PSF was used" )
		= withmirrorpsf;
		*rmfDS->addAttribute( "ANG_DIST", "ANGULAR_DIST extension included" )
		= withangdist;
	}
	else
	{
		rmfDS = dataSetServer->open( rmfset, Dal::Modify );

		source = rmfDS->attribute("SOURCEID")->asInt();
		order  = rmfDS->attribute("RFLORDER")->asInt();
		bkgcorrect = rmfDS->attribute("BKG_CORR")->asBool();
		withmirrorpsf = rmfDS->attribute("MIRRORPS")->asBool();
		withangdist = rmfDS->attribute("ANG_DIST")->asBool();

		betaGrid = gridspan( rmfDS->table("MATRIX")->column("F_CHAN") );
	}

	// source-dependent initializations

	string instrument = evenliDS->attribute("INSTRUME")->asString();
	calServer.setInstrument( Xmm::toInstrument(instrument) );

	//AI: Set the exposureTime to be used in the cal
	calServer.state()->exposureStartTime()->set( etab->attribute("DATE-OBS")->asString());

	calibInit( betaGrid, betaChannelGrid, betaORIChannelGrid,order);

	struct DataSet *srcliDS = dataSetServer->open( srclist, Dal::Read );

	float alpha, dpsi;
	init( srcliDS->table("SRCLIST"), source, alpha, dpsi, order );
	//AI: New call to calculate the vignetting to deal with the
	//lambda calculations. Only needed with lambda binning.
	//If we are dealing with beta, the vignetting are calculated within the
	//calibInit function.... FIXME!!!
	if (energyBinning == "lambda")
		calVignetting(betaChannelGrid,alpha);
	//AI: lambda binning

	char extname[32];
	struct ShapeScanner *foreground, *background, *pulseheight;

	std::sprintf( extname, "%s_SRC%d_SPATIAL", instrument.c_str(), source );
	foreground = scanner( srcliDS->table(extname), betaChannelGrid, xdspGrid ); //AI: Wavelength scale. Start

	if ( bkgcorrect ) {
		std::sprintf( extname, "%s_BACKGROUND", instrument.c_str() );
		background = scanner( srcliDS->table(extname), betaChannelGrid, xdspGrid ); //AI: Wavelength scale. Start
	}
	else
		background = 0;


	std::sprintf( extname, "%s_SRC%d_ORDER_%d", instrument.c_str(), source, order );
	pulseheight = scanner( srcliDS->table(extname), betaChannelGrid, Grid() );//AI: Wavelength scale. Start

	//AI: lambda binning order input parameter for lambda to beta
	// calculation
	init( lossByNode, evenliDS, *foreground, background, dpsi, order, energyBinning );
	init( banana, *pulseheight );

	delete pulseheight;
	delete background;
	delete foreground;

	dataSetServer->close( srcliDS );

	// construct the output dataset

	size_t channels = betaGrid.length();
	size_t betaChannels = betaChannelGrid.length(); //AI: Wavelength scale.
	unsigned currentRow;
	std::vector<double> enerMin,enerMax,enerMidPoints;
	if ( newrmf ) {
		// read the custom figure data if provided

		if ( withangdist ) {
			customFigure.clear();
			std::FILE *fi = std::fopen( angdistset.c_str(), "r" );
			if ( !fi )
				errstr << "could not open file " << angdistset
				<< Sas::error("FileNotFound");
			while ( !std::feof(fi) ) {
				char line[128];
				struct AngularDistBin b;
				if ( std::fgets(line,128,fi) &&
						3 == std::sscanf(line,"%f%f%f",&b.phi0,&b.phi1,&b.value) )
					customFigure.push_back(b);
			}
			std::fclose(fi);
		}

		// add the MATRIX extension

		struct Table *tp = rmfDS->addTable( "MATRIX", rows );

		//ARF
		if (witharffile)
		{
			struct Table *tpArf = arfDS->addTable( "SPECRESP", rows );
			tpArf->addAttribute( evenliDS->attribute("TELESCOP") );
			tpArf->addAttribute( evenliDS->attribute("INSTRUME") );

			*tpArf->addAttribute("FILTER") = "NONE";
			*tpArf->addAttribute("HDUCLASS") = "OGIP";
			*tpArf->addAttribute("HDUCLAS1") = "RESPONSE";
			*tpArf->addAttribute("HDUCLAS2","ARF extension")
			= "SPECRESP";
			*tpArf->addAttribute("HDUVERS") = "1.1.0";

			*tpArf->addAttribute("ARFVERSN","Obsolete") = "1992a";
			*tpArf->addAttribute("HDUVERS1","Obsolete") = "1.1.0";
			*tpArf->addAttribute("HDUVERS2","Obsolete") = "1.1.0";

			arfELCD = tpArf->addColumn( "ENERG_LO", Dal::Real32, "", "keV" )->data();
			arfELD = arfELCD->real32Data();
			arfEHCD = tpArf->addColumn( "ENERG_HI", Dal::Real32, "", "keV" )->data();
			arfEHD = arfEHCD->real32Data();
			arfCD = tpArf->addColumn( "SPECRESP", Dal::Real32,  "", "cm2" )->data();
			arfD = arfCD->real32Data();
		}
		//ARF

		tp->addAttribute( evenliDS->attribute("TELESCOP") );
		tp->addAttribute( evenliDS->attribute("INSTRUME") );

		*tp->addAttribute("FILTER") = "NONE";
		*tp->addAttribute("CHANTYPE") = "PI";
		*tp->addAttribute("DETCHANS") = (int)channels;
		*tp->addAttribute("HDUCLASS") = "OGIP";
		*tp->addAttribute("HDUCLAS1") = "RESPONSE";
		*tp->addAttribute("HDUCLAS2",bkgcorrect?"NET response":"TOTAL response")
		= "RSP_MATRIX";
		*tp->addAttribute("HDUCLAS3","all effects included") = "FULL";
		*tp->addAttribute("HDUVERS") = "1.3.0";
		*tp->addAttribute("PHAFILE","corresponding spectrum file")
		= withspectrum ? spectrumset : "NONE";
		*tp->addAttribute("LO_THRES","threshold of non-zero response")
		= low_threshold;

		*tp->addAttribute("RMFVERSN","Obsolete") = "1992a";
		*tp->addAttribute("HDUVERS1","Obsolete") = "1.1.0";
		*tp->addAttribute("HDUVERS2","Obsolete") = "1.2.0";

		tp->addColumn( "ENERG_LO", Dal::Real32, "", "keV" );
		tp->addColumn( "ENERG_HI", Dal::Real32, "", "keV" );
		tp->addColumn(    "N_GRP", Dal::Int16,  "", "" );

		tp->addColumn( "F_CHAN", Dal::Int16,  "", "", Dal::zero );
		tp->addColumn( "N_CHAN", Dal::Int16,  "", "", Dal::zero );
		tp->addColumn( "MATRIX", Dal::Real32, "", "", Dal::zero );

		struct Column *cp = tp->column("F_CHAN");
		*cp->addAttribute("TCTYP","channel axis type") = "";
		if (_lambdaSpec)
			*cp->addAttribute("TCUNI","channel axis units(-Angstrom*order)") = "Angstrom";
		else
			*cp->addAttribute("TCUNI","channel axis units") = "rad";
		*cp->addAttribute("TCRPX","reference channel number") = betaGrid.min;
		*cp->addAttribute("TCRVL","center of reference channel")
		= betaGrid[betaGrid.min];
		*cp->addAttribute("TCDLT","uniform channel width") = betaGrid.stride;
		*cp->addAttribute("TLMIN") = betaGrid.min;
		*cp->addAttribute("TLMAX") = betaGrid.max;

		// add the EBOUNDS extension

		tp = rmfDS->addTable( "EBOUNDS", channels );

		tp->addAttribute( evenliDS->attribute("TELESCOP") );
		tp->addAttribute( evenliDS->attribute("INSTRUME") );

		*tp->addAttribute("FILTER") = "NONE";
		*tp->addAttribute("CHANTYPE") = "PI";
		*tp->addAttribute("DETCHANS") = (int)channels;
		*tp->addAttribute("HDUCLASS") = "OGIP";
		*tp->addAttribute("HDUCLAS1") = "RESPONSE";
		*tp->addAttribute("HDUCLAS2") = "EBOUNDS";
		*tp->addAttribute("HDUVERS") = "1.2.0";
		*tp->addAttribute("PHAFILE","corresponding spectrum file")
		= withspectrum ? spectrumset : "NONE";

		*tp->addAttribute("RMFVERSN","Obsolete") = "1992a";
		*tp->addAttribute("HDUVERS1","Obsolete") = "1.0.0";
		*tp->addAttribute("HDUVERS2","Obsolete") = "1.1.0";

		DalVecI channel( tp, "CHANNEL", "", "channel" );
		DalVecE    eMin( tp,   "E_MIN", "", "keV" );
		DalVecE    eMax( tp,   "E_MAX", "", "keV" );

		for ( int i = betaGrid.min; i <= betaGrid.max; ++i ) {
			int j = i - betaGrid.min;
			//AI: Wavelength scale. Start

			if (_lambdaSpec)
			{
				eMin[j] = .001 * calLambdaEnergy(betaGrid[i+.5] / order);
				eMax[j] = .001 * calLambdaEnergy(betaGrid[i-.5] / order);
				enerMin.push_back(eMin[j]);
				enerMax.push_back(eMax[j]);
				//Energy bin mid point.
				//Used for effective area correction interpolation
				enerMidPoints.push_back(((eMax[j]+eMin[j])/2.0));
			}
			else
			{
				eMin[j] = .001 * calEnergy(betaGrid[i+.5],-order,alpha);
				eMax[j] = .001 * calEnergy(betaGrid[i-.5],-order,alpha);
			}

			//AI: Wavelength scale. End
			channel[j] = j + 1;
		}

		// add the ANGULAR_DIST extension if applicable

		if ( withangdist ) {
			tp = rmfDS->addTable( "ANGULAR_DIST", customFigure.size() );

			DalVecE  phi0( tp,  "PHI0", "minimum dispersion angle", "arcmin" );
			DalVecE  phi1( tp,  "PHI1", "maximum dispersion angle", "arcmin" );
			DalVecE value( tp, "VALUE", "distribution" );

			for ( unsigned i = 0; i < tp->rows(); ++i ) {
				const AngularDistBin &b = customFigure[i];
				phi0[i] = b.phi0;
				phi1[i] = b.phi1;
				value[i] = b.value;
			}
		}

		//

		currentRow = 0;
	}
	else { // --newrmf=false
		struct Table *tp;

		if ( withangdist ) {
			tp = rmfDS->table("ANGULAR_DIST");

			const DalVecE  phi0( tp->column("PHI0") );
			const DalVecE  phi1( tp->column("PHI1") );
			const DalVecE value( tp->column("VALUE") );

			customFigure.resize( tp->rows() );
			for ( unsigned i = 0; i < tp->rows(); ++i ) {
				struct AngularDistBin &b = customFigure[i];
				b.phi0 = phi0[i];
				b.phi1 = phi1[i];
				b.value = value[i];
			}
		}

		tp = rmfDS->table("MATRIX");

		struct Span replace;
		{
			const DalVecE energLo( tp->column("ENERG_LO") );
			const DalVecE energHi( tp->column("ENERG_HI") );

			for ( int i = 0; i+1 < (int)tp->rows(); ++i )
				if ( energLo[i] >= energHi[i] || energHi[i] != energLo[i+1] )
					errstr << "the incident energy bins "
					<< "are not contiguous in ascending order"
					<< Sas::error("InvalidMatrix");

			if ( emax <= energLo[0] ) {
				replace = Span(0).length(0);
				emax = energLo[0];
			}
			else if ( emin >= energHi[tp->rows()-1] ) {
				replace = Span(tp->rows()).length(0);
				emin = energHi[tp->rows()-1];
			}
			else {
				int i = tp->rows() - 1;

				while ( i > 0 && energLo[i] > emin ) --i;
				if ( i > 0 ) emin = energLo[i];
				replace.min = i;

				while ( i+1 < (int)tp->rows() && energHi[i] < emax ) ++i;
				if ( i+1 < (int)tp->rows() ) emax = energHi[i];
				replace.max = i;
			}
		}

		if ( rows < (int)replace.length() )
			tp->deleteRows( replace.min, replace.length() - rows );
		else if ( rows > (int)replace.length() )
			tp->insertRows( replace.min, rows - replace.length() );

		currentRow = replace.min;
	}

	//

	struct Table* table = rmfDS->table("MATRIX");

	struct Column* energLo = table->column("ENERG_LO");
	struct Column* energHi = table->column("ENERG_HI");
	struct Column*    nGrp = table->column("N_GRP");
	struct Column*   fChan = table->column("F_CHAN");
	struct Column*   nChan = table->column("N_CHAN");
	struct Column*  matrix = table->column("MATRIX");



	vector<float> lsf(betaChannels), accumulator(betaChannels),
	response(channels);

	vector<float> betaEffEner(rows);

	//Initialize response vector to 0;
	size_t j;
	for ( j = 0; j < channels; ++j ) response[j] = 0.;
	for ( int j = 0; j < rows; ++j ) betaEffEner[j] = 0.0;

	float lambda = hc/emin, deltaLambda = (hc/emin - hc/emax) / rows;


	if (witharffile)
	{
		for ( int i = 0; i < rows; ++i )
		{
			arfELD[i] = hc/lambda;
			lambda -= deltaLambda;
			arfEHD[i]  = hc/lambda;
		}
	}

	if  ( (withrectification == true) && (withEffAreaCorr == false))
	{
	  withEffAreaCorr = true;
	  errstr  << "withrectification parameter was set to YES. In this case, " <<
	      "witheffectiveareacorrection will be changed to YES automatically" <<
	      Sas::warning("RectificationDefaults");
	}
	if(withEffAreaCorr)
	{
		errstr << "Effective area correction (interpolation)." <<
				Sas::message(Msg::App,Msg::Verbose) ;
		interpolEffAreaCorr(enerMidPoints);
	}

	lambda = hc/emin, deltaLambda = (hc/emin - hc/emax) / rows;

	int messageTrigger = 1;
	float matrixSum = 0;
	for ( int rowCount = 1; rowCount <= rows; ++rowCount, ++currentRow ) {
		float eLo = DalVecE(energLo)[currentRow] = hc/lambda;
		lambda -= deltaLambda;
		float eHi = DalVecE(energHi)[currentRow] = hc/lambda;

		// small angles
		size_t i;
		for ( i = 0; i < betaChannels; ++i ) accumulator[i] = 0;

		for ( int ord = 0; ord < 6; ++ord )
		{
			prfNarrow(lsf,eLo, eHi, alpha, -ord, fftdim,
					withmirrorpsf, withangdist,energyBinning );
			for ( i = 0; i < betaChannels; ++i ) 	accumulator[i] += lsf[i];
		}

		nodeEfficiencySA( accumulator, (eLo+eHi)/2 );
		for ( i = 0; i < betaChannels; ++i ) response[i] = accumulator[i];

		// large angles

		for ( i = 0; i < betaChannels; ++i ) accumulator[i] = 0;

		for ( int ord = 0; ord < 6 ; ++ord ) {
			prfBroad( lsf,eLo, eHi, alpha, -ord, energyBinning );
			for ( i = 0; i < betaChannels; ++i ) accumulator[i] += lsf[i];
		}

		nodeEfficiencyLA( accumulator, dpsi, (eLo+eHi)/2 );

		for ( i = 0; i < betaChannels; ++i ) response[i] += accumulator[i];


		betaEfficiency( response, (eLo+eHi)/2. );

		if ( (withrectification == true) && (withEffAreaCorr == false))
                {
		  withEffAreaCorr = true;
		  errstr  << "withrectification parameter was set to YES. In this case, " <<
		    "witheffectiveareacorrection will be changed to YES automatically" <<
		    Sas::warning("RectificationDefaults");
		}
		if (withEffAreaCorr)
		{
			corrEffArea(response);
		}

		//Compute the effecitive area per input energy and
		//store the value in the final array
		if (witharffile)
		{
			float fracBE = 0.;
			for (unsigned i = 0 ; i < betaChannels ; ++i)
			{
				if ( response[i] >= low_threshold )
					fracBE += response[i];
			}

			betaEffEner[currentRow] = fracBE;

			//Normalization of the LSF

			double sum = 0.;
			for ( unsigned i = 0; i < betaChannels; ++i )
				sum += response[i];

			for ( unsigned i = 0; i < betaChannels; ++i )
				response[i] = response[i]*(1.0/sum);

		}


		struct CellData *fChanData = fChan->cellData(currentRow);
		struct CellData *nChanData = nChan->cellData(currentRow);
		struct CellData *matrixData= matrix->cellData(currentRow);

		vector< pair<unsigned,unsigned> > group;
		bool outside = true;
		unsigned nval = 0;


		for ( i = 0; i < response.size(); ++i )
		{
			if ( response[i] >= low_threshold )
			{
				if ( outside )
				{
					group.push_back( pair<unsigned,unsigned>(i,0) );
					outside = false;
				}
				++group.back().second;
				++nval;
			}
			else outside = true;
		}

		(DalVecI(nGrp)[currentRow] = group.size());

		fChanData->size(group.size());
		nChanData->size(group.size());
		matrixData->size(nval);

		float *dst = matrixData->real32Data();
		float kkresp = 0.0;
		for ( i = 0; i < group.size(); ++i )
		{
			(fChanData->int16Data())[i] = group[i].first + 1;
			int count = (nChanData->int16Data())[i] = group[i].second;
			vector<float>::const_iterator src = response.begin() + group[i].first;
			while ( count-- )
			{
				float kk =  *src++;
				kkresp = kkresp + kk;
				matrixSum += *dst++ = kk;
			}
		}

		delete matrixData;
		delete nChanData;
		delete fChanData;

		if ( messageTrigger <= 10. * rowCount / rows ) {
			errstr << "completed row " << currentRow+1
			<< ", current average effective area: " << matrixSum/rowCount
		<< Sas::message(Msg::App,Msg::Verbose);
			++messageTrigger;
		}

	}///rows LOOP

	if (witharffile)
	{
		for( unsigned int i = 0 ; i < betaEffEner.size() ; ++i)
			arfD[i] = betaEffEner[i];

		delete arfCD;
		delete arfEHCD;
		delete arfELCD;
	}

	dataSetServer->close( rmfDS );
	if (witharffile)
		dataSetServer->close( arfDS );
	dataSetServer->close( evenliDS );

	calibFree();

}

struct ShapeScanner* scanner( struct Table *table,
		const Grid &xGrid, const Grid &yGrid )
		{
	struct GeoRegion geoReg;
	struct InASCregionTable ascReg(table);
	ascReg >> geoReg;
	return new RegionScanner( geoReg, xGrid, yGrid );
		}

void init( const Table *table, int sourceid, float &alpha, float &dpsi, int order )
{
	const DalVecI index( table->column("INDEX") );

	for ( unsigned i = 0; i < table->rows(); ++i )
		if ( index[i] == sourceid ) {
			resetSourcePosition( DalVecE(table->column("FOV_R"))[i]*3600,
					DalVecE(table->column("FOV_PHI"))[i]*M_PI/180, order );
			alpha=calAlpha0()+DalVecE(table->column("DELTA_DISP"))[i]*M_PI/10800.;
			dpsi = DalVecE(table->column("DELTA_XDSP"))[i]*M_PI/10800;			
			return;
		}

	Sas::error("badSource","source is not present in SRCLIST table");
}

void LossStats::ingest( const ExposureMap &emap, const Spectrum &xpsf,
		int column, int offset )
{
	list<Span>::const_iterator spanItr;
	for ( spanItr = spans.begin(); spanItr != spans.end(); ++spanItr ) {
		struct Span span = Span(offset+spanItr->min,offset+spanItr->max);
		if ( !(span &= emap.bounds().row) ) continue;
		long itr = emap.id(column,span.min-1), end = emap.id(column,span.max);
		for ( int i = span.min - offset; itr != end; ) {
			float expsr = emap[ emap.incr(itr,Compass::North) ];
			crossPSF += xpsf[i++] * expsr;
			exposure += expsr;
		}
	}
}

void init( list<LossFunction> &lbn, struct DataSet *dataset,
		struct ShapeScanner &fgScanner, struct ShapeScanner *bgScanner,
		float dpsi, int order, string energyBinning )
{
	lbn.clear();

	struct CanonicalCrossPsf *crossPSF = 0;
	crossPSF = calServer.getAtom(crossPSF);

	CalReal64Vector xdspBins( 0, xdspGrid.length() );
	for ( int i = xdspGrid.min; i <= xdspGrid.max; ++i )
		xdspBins[i-xdspGrid.min] = xdspGrid[i];

	//

	unsigned validCount =  0;
	for ( unsigned ccd = 1; ccd <= 9; ++ccd )
		for ( unsigned node = 0; node < 2; ++node ) {
			char extname[9];
			std::sprintf( extname, "EXPMAP%u%u", node, ccd );
			if ( !dataset->hasBlock(extname) ) continue;
			struct ExposureMap emap( dataset->array(extname) );


			bool valid = true;
			if ( !subdivides(GridSpan(emap.beta,emap.bounds().col),betaChannelGrid) ) { //AI: Wavelength scale. betaGrid Changed!!
				errstr << "the exposure map dispersion channels are not "
				<< "compatible with the EBOUNDS table";
				valid = false;
			}
			if ( !commensurate(emap.xdsp,xdspGrid) ) {
				errstr << "the exposure map cross-dispersion channels are not "
				<< "commensurate with the EVENTS table";
				valid = false;
			}
			if ( !emap.bounds() ) {
				errstr << "the exposure map is null";
				valid = false;
			}

			if ( !valid ) {
				errstr << ": discarding " << extname << Sas::warning("invalidMap");
				continue;
			}
			else ++validCount;

			// initialize the clipping region

			struct Bounds clip;
			clip.col = Span( betaChannelGrid.cell( emap.beta[emap.bounds().col.min] ), //AI: Wavelength scale. Changed
					betaChannelGrid.cell( emap.beta[emap.bounds().col.max] ) ); //AI: Wavelength scale. Changed
			clip.row = Span( xdspGrid.cell( emap.xdsp[emap.bounds().row.min] ),
					xdspGrid.cell( emap.xdsp[emap.bounds().row.max] ) );
			clip &= Bounds(betaChannelGrid,xdspGrid); //AI: Wavelength scale. Changed

			// initialize the loss function

			size_t size = clip.col.length();

			lbn.push_back( LossFunction() );
			struct LossFunction &loss = lbn.back();
			loss.ccd = ccd;
			loss.offset = clip.col.min - betaChannelGrid.min; //AI: Wavelength scale. Changed

			// initialize the accumulators

			list<Span> spans;
			int b;

			loss.fg.resize(size);
			fgScanner.reset(&clip);
			while ( fgScanner.next(b,spans) )
				loss.fg[b-clip.col.min].spans = spans;

			if ( bgScanner ) {
				loss.bg.resize(size);
				bgScanner->reset(&clip);
				while ( bgScanner->next(b,spans) )
					loss.bg[b-clip.col.min].spans = spans;
			}

			// accumulate over this node

			const Bounds &bnds = emap.bounds();
			int offset = bnds.row.min - xdspGrid.cell( emap.xdsp[bnds.row.min] );
			int prev = clip.col.min - 1;
			struct Spectrum dist;
			//AI: Wavelength scale. Start
			//double cosBeta,realBeta;
			offaxis = 0.0;
			if (dataset->table("EVENTS")->hasAttribute("OFFAXIS"))
				offaxis =  dataset->table("EVENTS")->attribute("OFFAXIS")->asReal();
			struct MiscDataServer *misc = 0; misc = calServer.getAtom(misc);
			alpha0 = (misc->get("INCIDENCE_ANGLE")
					+ misc->get("LINCOORD:LINCOORD%G_RY")) * M_PI/180.;
			linespacing = misc->get("GRAT_LINE_DENS");
			calServer.releaseAtom(misc);
			//AI: Wavelength scale. End

			for ( int column = bnds.col.min; column <= bnds.col.max; ++column ) {
				//In stead of converting the lambda value, we will convert the energy to
				//beta corr values.???
				b = betaChannelGrid.cell( emap.beta[column] );//AI: Wavelength scale. Changed

				if ( b < clip.col.min ) continue;
				else if ( b > clip.col.max ) break;

				if ( b != prev ) {
					//AI: lambda binning
					// lambda to rad...
					double betaVal;
					if (energyBinning == "lambda")
						betaVal = acos((betaChannelGrid[b]*(-order)/linespacing) + std::cos(double(alpha0+offaxis)));
					else
						betaVal = betaChannelGrid[b];

					dist = crossPSF->probability(double(0.0),betaVal,double(dpsi), //AI: Wavelength scale. Changed
							xdspBins,xdspGrid.stride);

					//AI: lambda binning
					prev = b;
				}

				loss.fg[b-clip.col.min].ingest( emap, dist, column, offset );
				if ( bgScanner )
					loss.bg[b-clip.col.min].ingest( emap, dist, column, offset );
			}

			// normalize the stats

			float datamax = emap.array->attribute("DATAMAX")->asReal();
			float normalizer = emap.beta.stride / betaChannelGrid.stride / datamax;  //AI: Wavelength scale. Changed

			vector<LossStats>::iterator itr;
			for ( itr = loss.fg.begin(); itr != loss.fg.end(); ++itr )
				itr->normalize( normalizer );

			if ( bgScanner )
				for ( itr = loss.bg.begin(); itr != loss.bg.end(); ++itr )
					itr->normalize( normalizer );
		}

	if ( validCount == 0 )
		Sas::error("nullExposure","no valid exposure maps are present");

	calServer.releaseAtom(crossPSF);

}

void init( vector< list<Span> > &banana, struct ShapeScanner &scanner )
{
	banana.clear();
	banana.resize( betaChannelGrid.length() );//AI: Wavelength scale. Changed

	int b;
	list<Span> spans;
	struct Bounds clip( betaChannelGrid, Span(0,scanner.bounds().row.max) );//AI: Wavelength scale. Changed
	scanner.reset( &clip );

	while ( scanner.next(b,spans) ) banana[b-betaChannelGrid.min] = spans; //AI: Wavelength scale. Changed
}

void nodeEfficiencySA( vector<float> &rf, float energy )
{
	vector<float> lossf( betaChannelGrid.length() ); //AI: Wavelength scale. Changed

	float qe[10];
	ccd_qe( qe, energy );

	list<LossFunction>::const_iterator itr;
	for ( itr = lossByNode.begin(); itr != lossByNode.end(); ++itr ) {
		size_t b = itr->offset;
		const float ccdqe = qe[ itr->ccd ];
		vector<LossStats>::const_iterator fgI, bgI = itr->bg.begin();
		for ( fgI = itr->fg.begin(); fgI != itr->fg.end(); ++fgI ) {
			if ( b >= lossf.size() ) break; // necessary?
			float xpsf = fgI->crossPSF;
			if ( bgI != itr->bg.end() ) {
				float scale = bgI->exposure ? fgI->exposure / bgI->exposure : 0;
				xpsf -= scale * (bgI++)->crossPSF;
			}
			lossf[b++] += xpsf*ccdqe ;
		}
	}


	for ( size_t i = 0; i < lossf.size(); ++i )
	{
		// In some detector bins (6.365, 6.375, 9.095, 15.365..., 31.205 Angstroms)
		// the ccdqe is counted twice.
		if (lossf[i] > 1.0)
			lossf[i] = lossf[i]/2.;

		rf[i] *= lossf[i];
	}

}

void nodeEfficiencyLA( vector<float> &rf, float dpsi, float energy)
{

	vector<float> lossf( betaChannelGrid.length() ); //AI: Wavelength scale. Changed


	float qe[10];

	ccd_qe( qe, energy );

	list<LossFunction>::const_iterator itr;
	for ( itr = lossByNode.begin(); itr != lossByNode.end(); ++itr ) {
		size_t b = itr->offset;
		const float ccdqe = qe[ itr->ccd ];
		vector<LossStats>::const_iterator fgI, bgI = itr->bg.begin();
		for ( fgI = itr->fg.begin(); fgI != itr->fg.end(); ++fgI ) {
			if ( b >= lossf.size() ) break; // necessary?

			float xpsf = empiricalCrossLA( energy, dpsi, fgI->spans );
			if ( bgI != itr->bg.end() )
				xpsf -= empiricalCrossLA( energy, dpsi, (bgI++)->spans );

			lossf[b++] += fgI->exposure * xpsf*ccdqe ;
		}
	}


	for ( size_t i = 0; i < lossf.size(); ++i )
	{
		// In some detector bins (6.365, 6.375, 9.095, 15.365..., 31.205 Angstroms)
		// the ccdqe is counted twice.
		if (lossf[i] > 1.0)
			lossf[i] = lossf[i]/2.;

		rf[i] *= lossf[i];
	}
}

void betaEfficiency( vector<float> &rf, float energy)
{
	const float energy_ph_tolerance = 0.01; // keV

	static vector<float> lossf( betaChannelGrid.length() );	 //AI: Wavelength scale. Changed

	static float energy0;
	size_t i;

	if ( !betaEfficInit ) {
		energy0 = energy - 2 * energy_ph_tolerance;
		betaEfficInit = true;
	}

	//if ( fabs(energy0 - energy) > energy_ph_tolerance ) {
		energy0 = energy + (energy > energy0 ? 1 : -1) * energy_ph_tolerance;

		calServer.state()->ccd()->set(1); // any CCD is fine
		struct CanonicalRedist *redist = 0; redist = calServer.getAtom(redist);
		struct Spectrum dist( redist->spectrum(1000.*energy) );

		calServer.releaseAtom(redist);

		float sum = 0;
		vector<float> cdist( dist.size() );
		for ( i = 0; i < dist.size(); ++i )	cdist[i] = (sum += dist[i]);


		list<Span>::const_iterator itr;
		struct Span s, distSpan = Span(0).length(dist.size());

		for ( i = 0; i < lossf.size(); ++i ) {
			lossf[i] = 0;
			for ( itr = banana[i].begin(); itr != banana[i].end(); ++itr )
				if ( (s = *itr & distSpan) )
					lossf[i] += cdist[s.max] - (s.min<1 ? 0 : cdist[s.min-1]);
		}
	//}

		for ( i = 0; i < rf.size(); ++i ) rf[i] *=  vignetting[i] * lossf[i];

}

float empiricalCrossLA( float energy, float dpsi, const list<Span> &spans )
{
	int n = 0;
	float s = 0;
	list<Span>::const_iterator itr;
	for ( itr = spans.begin(); itr != spans.end(); ++itr ) {
		s += empiricalCrossLA( energy, dpsi,
				xdspGrid[itr->min-.5], xdspGrid[itr->max+.5] );
		n += itr->length();
	}
	return s/n;
}


bool multiple( float multiple, float divisor, float scale )
{
	float count = floor( 0.5 + multiple / divisor );
	return fabs(multiple - count * divisor) <= 5e-7 * fabs(scale);
}

bool subdivides( const GridSpan &fine, const GridSpan &coarse )
{
	return
	multiple( coarse.stride, fine.stride, coarse.stride )
	&&
	multiple( fine.minval() - coarse.minval(), fine.stride,
			max(fabs(fine.origin),fabs(coarse.origin)) );
}

